"""A game very similar to Simon, but backwards."""
import enum
from typing import Any, Dict, Optional, Tuple

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType

from popgym.core.deck import Deck
from popgym.core.env import POPGymEnv


class Mode(enum.IntEnum):
    PLAY = 0
    WATCH = 1


class Autoencode(POPGymEnv):
    """A game very similar to Simon, but backwards.

    The agent receives a sequence of cards, and must output the cards it saw
    in reverse order. E.g., seeing [1, 2, 3] means I should output them in the order
    [1, 2, 3].

    Examples:
        >>> env = Autoencode()
        >>> env.reset()

    Args:
        num_decks: The maximum number of decks the agent must memorize

    Returns:
        A gym environment
    """

    def __init__(self, num_decks=1):
        self.deck = Deck(num_decks)
        self.deck.add_players("system")
        self.max_episode_length = self.deck.num_cards * 2 - 1
        self.action_space = self.deck.get_obs_space(["suits"])
        self.observation_space = gym.spaces.Tuple(
            (
                gym.spaces.Discrete(2),
                self.action_space,
            )
        )
        self.state_space = gym.spaces.Tuple(
            (
                gym.spaces.MultiDiscrete([4] * self.deck.num_cards),
                gym.spaces.Discrete(2),
                gym.spaces.Box(0, 1, (1,)),
            )
        )
        self.mode = Mode.WATCH

    def make_obs(self, card_idx) -> Tuple[int, int]:
        card_suit = self.deck.suits_idx[card_idx].reshape(-1)
        return int(self.mode.value), card_suit.item()

    def get_state(self) -> Tuple[np.ndarray, int, np.ndarray]:
        cards = self.deck.suits_idx[self.deck.idx].copy()
        mode = int(self.mode.value)
        pos = np.array(
            [len(self.deck["system"]) / self.deck.num_cards], dtype=np.float32
        )
        return cards, mode, pos

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        terminated = truncated = False
        reward = 0
        # TODO: This is causing flaky tests, make sure we are not
        # off-by-one in step
        reward_scale = 1 / self.deck.num_cards

        if self.mode == Mode.WATCH:
            self.deck.deal("system", 1)
            if len(self.deck) == 0:
                self.mode = Mode.PLAY
                # Flip the cards so they play backwards
                # self.deck["system"].reverse()
            obs = self.make_obs(self.deck["system"][-1])
        else:
            # Recited all cards
            terminated = len(self.deck["system"]) == 1
            correct_card = self.deck.suits_idx[self.deck["system"].pop(-1)]
            if action == correct_card:
                reward = reward_scale
            else:
                reward = -reward_scale

            obs = self.make_obs(np.array([0]))

        info: dict = {}

        return obs, reward, terminated, truncated, info

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[gym.core.ObsType, Dict[str, Any]]:
        super().reset(seed=seed)
        self.deck.reset(rng=self.np_random)
        self.deck.deal("system", 1)

        self.mode = Mode.WATCH
        obs = self.make_obs(self.deck["system"][-1])
        return obs, {}


class AutoencodeEasy(Autoencode):
    pass


class AutoencodeMedium(Autoencode):
    def __init__(self):
        super().__init__(num_decks=2)


class AutoencodeHard(Autoencode):
    def __init__(self):
        super().__init__(num_decks=3)



import numpy as np

from popgym.envs.autoencode import Autoencode, Mode
from tests.base_env_test import AbstractTest


class TestAutoencode(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = Autoencode()

    def test_full(self):
        obs, _ = self.env.reset()
        terminated = truncated = False
        seq = [obs[1]]
        reward_sum = 0
        while obs[0] == Mode.WATCH:
            obs, reward, terminated, truncated, _ = self.env.step(0)
            seq.append(np.array(obs[1]))
            reward_sum += reward
        self.assertFalse(terminated or truncated)
        self.assertEqual(reward_sum, 0)
        for i in range(len(seq)):
            self.assertFalse(terminated or truncated)
            a = seq.pop(-1)
            _, reward, terminated, truncated, _ = self.env.step(a)
            reward_sum += reward
        self.assertTrue(terminated or truncated)
        self.assertEqual(len(self.env.deck["system"]), 0)

        self.assertAlmostEqual(reward_sum, 1.0)



import numpy as np

from popgym.envs.battleship import Battleship
from tests.base_env_test import AbstractTest


class TestBattleship(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = Battleship()

    def test_hit_miss(self):
        self.env.reset()
        ship_positions = np.where(self.env.board)  # (x_pos_arr, y_pos_arr)
        hit_pos = (ship_positions[0][0], ship_positions[1][0])
        no_ship_positions = np.where(~self.env.board.astype(bool))
        miss_pos = (no_ship_positions[0][0], no_ship_positions[1][0])
        obs1, reward1, terminated1, truncated1, info1 = self.env.step(
            hit_pos
        )  # test hit
        assert obs1 == 1
        assert reward1 == 1.0 / sum(self.env.ship_sizes)
        obs2, reward2, terminated2, truncated2, info2 = self.env.step(
            hit_pos
        )  # test duplicate hit position (should be a miss now)
        assert obs2 == 0
        assert reward2 == -1.0 / (
            self.env.max_episode_length - sum(self.env.ship_sizes)
        )
        obs3, reward3, terminated3, truncated3, info3 = self.env.step(
            miss_pos
        )  # test miss
        assert obs3 == 0
        assert reward3 == -1.0 / (
            self.env.max_episode_length - sum(self.env.ship_sizes)
        )

    def test_num_ships(self):
        self.env.reset()
        assert sum(self.env.ship_sizes) == np.sum(self.env.board)

    def test_reward_sum(self):
        self.env.reset()

        # Episode reward 1.0 (only guesses hits)
        reward = 0.0
        ship_positions = np.where(self.env.board)  # (x_pos_arr, y_pos_arr)
        for i in range(len(ship_positions[0])):
            hit_pos = (ship_positions[0][i], ship_positions[1][i])
            obsi, rewardi, terminatedi, truncatedi, infoi = self.env.step(hit_pos)
            reward += rewardi
            if i < len(ship_positions[0]) - 1:
                assert not (terminatedi or truncatedi)
            elif i == len(ship_positions[0]) - 1:
                assert terminatedi
        self.assertAlmostEqual(reward, 1.0)

        # Episode reward minimum
        self.env.reset()
        reward = 0.0
        no_ship_positions = np.where(~self.env.board.astype(bool))
        for i in range(self.env.max_episode_length):
            miss_pos = (
                no_ship_positions[0][i % len(no_ship_positions)],
                no_ship_positions[1][i % len(no_ship_positions)],
            )
            obsi, rewardi, terminatedi, truncatedi, infoi = self.env.step(miss_pos)
            reward += rewardi
            if i < self.env.max_episode_length - 1:
                assert not (terminatedi or truncatedi)
            elif i == self.env.max_episode_length - 1:
                assert truncatedi
        self.assertAlmostEqual(
            reward, -1.0 * self.env.max_episode_length / len(no_ship_positions[0])
        )

        # Episode reward 0.0
        self.env.reset()
        reward = 0.0
        for x in range(self.env.board_size):
            for y in range(self.env.board_size):
                obsi, rewardi, terminatedi, truncatedi, infoi = self.env.step((x, y))
                reward += rewardi
        self.assertAlmostEqual(reward, 0.0)



from typing import List, Tuple

import gymnasium as gym
import torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType
from torch import nn

from popgym.baselines.ray_models.base_model import BaseModel


class LSTM(BaseModel):
    r"""Long Short-Term Memory from

    .. code-block:: text

        @article{hochreiter_long_1997,
            title = {Long Short-Term Memory},
            volume = {9},
            issn = {0899-7667},
            url = {https://doi.org/10.1162/neco.1997.9.8.1735},
            doi = {10.1162/neco.1997.9.8.1735},
            number = {8},
            journal = {Neural Comput.},
            author = {Hochreiter, Sepp and Schmidhuber, Jürgen},
            month = nov,
            year = {1997},
            pages = {1735--1780},
        }
    """

    MODEL_CONFIG = {
        # Number of recurrent hidden layers in encoder/decoder
        "num_recurrent_layers": 1,
        "benchmark": False,
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        # Need to define self.core
        if self.cfg["benchmark"]:
            self.core = nn.LSTMCell(
                self.cfg["preprocessor_output_size"],
                self.cfg["hidden_size"],
            )
        else:
            self.core = nn.LSTM(
                self.cfg["preprocessor_output_size"],
                self.cfg["hidden_size"],
                self.cfg["num_recurrent_layers"],
                batch_first=True,
            )

    def initial_state(self) -> List[TensorType]:
        return [
            torch.zeros(1, self.cfg["hidden_size"]),
            torch.zeros(1, self.cfg["hidden_size"]),
        ]

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:

        state = [s.permute(1, 0, 2) for s in state]

        if self.cfg["benchmark"]:
            outs = []
            state = [s.squeeze(0) for s in state]
            for t in range(z.shape[1]):
                state = self.core(z[:, t], state)
                outs.append(state[0])
            z = torch.stack(outs, dim=1)
            state = [s.unsqueeze(0) for s in state]
        else:
            z, state = self.core(z, state)

        state = [s.permute(1, 0, 2) for s in state]
        return z, state



from typing import List, Tuple

import gymnasium as gym
import torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType

from popgym.baselines.models.lmu import LMU as LMUModel
from popgym.baselines.ray_models.base_model import BaseModel


class LMU(BaseModel):
    r"""Legendre Memory Units from

    .. code-block:: text

        @inproceedings{voelker_legendre_2019,
            title = {
                Legendre Memory Units: Continuous-Time
                Representation in Recurrent Neural Networks
            },
            volume = {32},
            shorttitle = {Legendre {Memory} {Units}},
            urldate = {2022-09-22},
            booktitle = {Advances in {Neural} {Information} {Processing} {Systems}},
            publisher = {Curran Associates, Inc.},
            author = {Voelker, Aaron and Kajić, Ivana and Eliasmith, Chris},
            year = {2019},
        }
    """

    MODEL_CONFIG = {
        "embedding": None,
        "theta": 64,
        "learn_a": False,
        "learn_b": False,
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        self.core = LMUModel(
            input_size=self.cfg["preprocessor_output_size"],
            hidden_size=self.cfg["hidden_size"],
            memory_size=self.cfg["hidden_size"],
            theta=self.cfg["theta"],
            learn_a=self.cfg["learn_a"],
            learn_b=self.cfg["learn_b"],
        )

    def initial_state(self) -> List[TensorType]:
        return [
            torch.zeros(1, self.cfg["hidden_size"]),
            torch.zeros(1, self.cfg["hidden_size"]),
        ]

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:

        hidden, memory = state

        z, (hidden, memory) = self.core(z, (hidden.squeeze(1), memory.squeeze(1)))

        # State expected to be list
        return z, [hidden.unsqueeze(1), memory.unsqueeze(1)]



from typing import Any, Dict, Optional, Tuple

import gymnasium as gym
from gymnasium.core import ActType, ObsType

from popgym.core.maze import Actions, Cell, MazeEnv


class LabyrinthEscape(MazeEnv):
    """A maze environment where the agent receives negative rewards
     until it finds the goal.

    Args:
        maze_dims: (heigh, width) of the generated mazes in blocks
        episode_length: maximum length of an episode

    Returns:
        A gym environment

    """

    def __init__(self, maze_dims=(10, 10), episode_length=1024):
        super().__init__(maze_dims, episode_length)
        self.neg_reward_scale = -1 / self.max_episode_length

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        super().step(action)
        reward = self.neg_reward_scale
        terminated = False
        y, x = self.position
        if self.maze.grid[y, x] == Cell.GOAL:
            reward += 1.0
            terminated = True
        truncated = self.curr_step == self.max_episode_length - 1

        obs = self.get_obs(action)
        info = {"position": (x, y)}

        return obs, reward, terminated, truncated, info

    def render(self):
        print(self.tostring(start=True, end=True, agent=True, visited=True))

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[gym.core.ObsType, Dict[str, Any]]:
        super().reset(seed=seed)
        # Based on free space
        x, y = self.position
        self.maze.grid[self.maze.end] = Cell.GOAL
        obs = self.get_obs(Actions.NONE)

        return obs, {"maze": str(self.maze.grid)}


class LabyrinthEscapeEasy(LabyrinthEscape):
    def __init__(self):
        super().__init__(maze_dims=(10, 10), episode_length=1024)


class LabyrinthEscapeMedium(LabyrinthEscape):
    def __init__(self):
        super().__init__(maze_dims=(14, 14), episode_length=1024)


class LabyrinthEscapeHard(LabyrinthEscape):
    def __init__(self):
        super().__init__(maze_dims=(18, 18), episode_length=1024)



from typing import List, Tuple

import gymnasium as gym
import torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType

from popgym.baselines.models.indrnn import IndRNN as IndRNNModel
from popgym.baselines.ray_models.base_model import BaseModel


class IndRNN(BaseModel):
    r"""A two-layer independently recurrent neural networks from

    .. code-block:: text

        @inproceedings{li_independently_2018,
            address = {Salt Lake City, UT},
            title = {
                Independently Recurrent Neural Network (IndRNN):
                Building a Longer and Deeper RNN
            },
            isbn = {978-1-5386-6420-9},
            shorttitle = {Independently {Recurrent} {Neural} {Network} ({IndRNN})},
            url = {https://ieeexplore.ieee.org/document/8578670/},
            doi = {10.1109/CVPR.2018.00572},
            language = {en},
            urldate = {2022-09-21},
            booktitle = {
                2018 {IEEE}/{CVF} {Conference} on {Computer} {Vision}
                and {Pattern} {Recognition}
            },
            publisher = {IEEE},
            author = {
                Li, Shuai and Li, Wanqing and Cook, Chris and Zhu, Ce and Gao, Yanbo
            },
            month = jun,
            year = {2018},
            pages = {5457--5466},
        }
    """

    MODEL_CONFIG = {
        "activation": "tanh",
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        self.core = IndRNNModel(
            input_size=self.cfg["preprocessor_output_size"],
            hidden_size=self.cfg["hidden_size"],
            max_len=model_config["max_seq_len"],
        )

    def initial_state(self) -> List[TensorType]:
        return [
            torch.zeros(2, self.cfg["hidden_size"]),
        ]

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:
        B, T, _ = z.shape
        [memory] = state
        z, memory = self.core(z, memory)
        return z, [memory]



from gettext import npgettext
from typing import Tuple
import gymnasium as gym
from gymnasium.core import ActType, ObsType
from popgym.core.env import POPGymEnv
from popgym.core.wrapper import POPGymWrapper
import numpy as np

class Flatten(POPGymWrapper):
    """
    Wrapper that flattens the observation and action spaces
    to make them compatible with neural networks.
    """

    def __init__(self, env: POPGymEnv, flatten_action: bool=True, flatten_observation: bool=True):
        super().__init__(env)
        self.need_flatten_action = False
        self.need_flatten_obs = False
        if flatten_action:
            # Check that spaces are either all continuous or all discrete
            self.continuous(self.env.action_space)
            self.need_flatten_action = isinstance(self.env.action_space, gym.spaces.Tuple) or isinstance(self.env.action_space, gym.spaces.Dict)
            if self.need_flatten_action:
                self.action_space = gym.spaces.utils.flatten_space(self.env.action_space)
        if flatten_observation:
            self.continuous(self.env.observation_space)
            self.need_flatten_obs = isinstance(self.env.action_space, gym.spaces.Tuple) or isinstance(self.env.action_space, gym.spaces.Dict)
            if self.need_flatten_obs:
                self.observation_space = gym.spaces.utils.flatten_space(self.env.observation_space)

    
    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        if self.need_flatten_action:
            flat_action = gym.spaces.utils.unflatten(self.env.action_space, action)
        else:
            flat_action = action

        obs, reward, terminated, truncated, info = self.env.step(flat_action)

        if self.need_flatten_obs:
            obs = gym.spaces.utils.flatten(self.env.observation_space, obs)
        return obs, reward, terminated, truncated, info

    def reset(self, **kwargs):
        obs, info = self.env.reset(**kwargs)
        if self.need_flatten_obs:
            obs = gym.spaces.utils.flatten(self.env.observation_space, obs)
        return obs, info

    def continuous(self, space: gym.spaces.Space) -> bool:
        if isinstance(space, gym.spaces.Box):
            if np.issubdtype(space.dtype, np.integer):
                return False
            else:
                return True
        elif isinstance(space, (gym.spaces.Discrete, gym.spaces.MultiDiscrete, gym.spaces.MultiBinary)):
            return False
        elif isinstance(space, gym.spaces.Tuple):
            res = [self.continuous(s) for s in space.spaces]
            assert all(res) or not any(res), "Cannot mix continuous and discrete spaces"
            return any(res)
        elif isinstance(space, gym.spaces.Dict):
            res = [self.continuous(s) for s in space.spaces.values()]
            assert all(res) or not any(res), "Cannot mix continuous and discrete spaces"
            return any(res)
        else:
            raise NotImplementedError(f"Unknown space: {space}")


from popgym.envs.multiarmed_bandit import MultiarmedBandit
from tests.base_env_test import AbstractTest


class TestBandits(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = MultiarmedBandit()



"""Classic game of Battleship, except the player's board is hidden"""
from typing import Any, Dict, Optional, Tuple

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType

from popgym.core.env import POPGymEnv


class Battleship(POPGymEnv):
    """Classic game of Battleship, except the board is obscured. Instead, the agent
    receives a hit/miss notification with the tile coordinates.

    Args:
        board_size: The width/height of the board in tiles
        ship_sizes: A list of ships sizes (in tiles) to place on the board

    Returns:
        A gym environment
    """

    obs_requires_prev_action = True

    def __init__(self, board_size=10, ship_sizes=[2, 3, 3, 4]):
        # Params
        self.board_size = board_size
        self.ship_sizes = ship_sizes
        self.max_episode_length = self.board_size**2
        self.observation_space = gym.spaces.Discrete(2)
        self.state_space = gym.spaces.Tuple(
            (
                gym.spaces.MultiDiscrete(
                    np.full((self.board_size, self.board_size), 2, dtype=int)
                ),
                gym.spaces.MultiDiscrete(
                    np.full((self.board_size, self.board_size), 2, dtype=int)
                ),
                self.observation_space,
            )
        )
        self.action_space = gym.spaces.MultiDiscrete([self.board_size, self.board_size])
        self.last_obs = 0

    def get_state(self):
        state = (self.board.copy(), self.guesses.copy(), self.last_obs)
        return state

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        hit = False
        is_ship = self.board[action[0], action[1]]
        guessed_before = self.guesses[action[0], action[1]]
        self.guesses[action[0], action[1]] = 1
        if is_ship and not guessed_before:
            hit = True
            self.hits += 1

        self.num_steps += 1

        # Episode finishes when the max steps is reached
        # (which defaults to the number of squares on the board)
        #       or when all of the ships have been sunk
        truncated = self.num_steps >= self.max_episode_length
        terminated = self.hits == self.needed_hits
        # Obs is 1 for a hit and 0 for a miss
        obs = int(hit)
        # Reward is structured so that an episode of all hits will have a
        # total reward of 1.0,
        #
        # an episode with all misses will have a total reward of
        # -1.0 * (max_steps / (max_steps-num_hits_needed)),
        #
        # and an episode which simply guesses every single square on the
        # board will have a reward of 0.0
        reward = (int(hit) * (1.0 / self.needed_hits)) + (
            int(not hit) * (-1.0 / (self.max_episode_length - self.needed_hits))
        )
        info: dict = {}
        self.last_obs = obs

        return obs, reward, terminated, truncated, info

    def place_ship(self, board, ship_size):
        valid = False
        start = None
        end = None
        while not valid:
            idx_start = self.np_random.integers(self.board_size, size=2)
            direction = self.np_random.integers(2, size=2)
            idx_end = idx_start.copy()
            idx_end[direction[0]] = idx_start[direction[0]] + (-1) ** direction[1] * (
                ship_size - 1
            )
            if np.any(idx_end < 0) or np.any(idx_end >= self.board_size):
                continue
            start = np.minimum(idx_start, idx_end)
            end = np.maximum(idx_start, idx_end)
            valid = np.sum(board[start[0] : end[0] + 1, start[1] : end[1] + 1]) == 0
        board[start[0] : end[0] + 1, start[1] : end[1] + 1] = 1

        return board

    def render(self, mode=None):
        viz = np.chararray(self.board.shape, unicode=True)
        viz[self.board == 0] = "·"
        viz[self.board == 1] = "☐"
        viz[(self.guesses == 1) * (self.board == 1)] = "☒"
        viz[(self.guesses == 1) * (self.board == 0)] = "⚬"

        print(" " + str(viz).replace("[", "").replace("]", "").replace("'", ""))

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[gym.core.ObsType, Dict[str, Any]]:

        super(Battleship, self).reset(seed=seed)
        self.num_steps = 0

        self.board = np.zeros((self.board_size, self.board_size), dtype=np.int8)
        for ship_size in self.ship_sizes:
            self.board = self.place_ship(self.board, ship_size)

        self.guesses = np.zeros((self.board_size, self.board_size), dtype=np.int8)

        self.hits = 0
        obs = 0
        self.last_obs = obs

        self.needed_hits = sum(self.ship_sizes)
        return obs, {}


class BattleshipEasy(Battleship):
    def __init__(self):
        super().__init__(board_size=8, ship_sizes=[2, 3, 3, 4])


class BattleshipMedium(Battleship):
    def __init__(self):
        super().__init__(board_size=10, ship_sizes=[2, 3, 3, 4])


class BattleshipHard(Battleship):
    def __init__(self):
        super().__init__(board_size=12, ship_sizes=[2, 3, 3, 4])



from typing import Optional, Tuple

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType

from popgym.envs.stateless_cartpole import StatelessCartPole


class NoisyStatelessCartPole(StatelessCartPole):
    def __init__(self, *args, **kwargs):
        noise_sigma = kwargs.pop("noise_sigma", 0.1)
        super().__init__(*args, **kwargs)
        self.noise_sigma = np.full(self.observation_space.shape, noise_sigma)

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        next_obs, reward, terminated, truncated, info = super().step(action)
        next_obs = self.add_noise_to_obs(next_obs)
        return next_obs, reward, terminated, truncated, info

    def add_noise_to_obs(self, obs):
        noise = self.np_random.normal(0, self.noise_sigma).astype(np.float32)
        obs_space = self.observation_space
        obs = np.clip(obs + noise, obs_space.low, obs_space.high)
        return obs

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[dict] = None
    ) -> Tuple[gym.core.ObsType, dict]:

        init_obs, info = super().reset(seed=seed, options=options)
        init_obs = self.add_noise_to_obs(init_obs)
        return init_obs, info


class NoisyStatelessCartPoleEasy(NoisyStatelessCartPole):
    def __init__(self):
        super().__init__(noise_sigma=0.1)


class NoisyStatelessCartPoleMedium(NoisyStatelessCartPole):
    def __init__(self):
        super().__init__(noise_sigma=0.2)


class NoisyStatelessCartPoleHard(NoisyStatelessCartPole):
    def __init__(self):
        super().__init__(noise_sigma=0.3)



import unittest

import numpy as np

from popgym.envs import has_mazelib

if has_mazelib():
    from popgym.envs.labyrinth_explore import LabyrinthExplore

    class TestLabyrinthExplore(unittest.TestCase):
        def test_step(self):
            e = LabyrinthExplore()
            obs, _ = e.reset()
            self.assertEqual(obs.dtype, np.int32)
            terminated = truncated = False
            for i in range(1000):
                obs, _, terminated, truncated, _ = e.step(e.action_space.sample())
                self.assertEqual(obs.dtype, np.int32)
                if terminated or truncated:
                    obs = e.reset()
                    self.assertEqual(obs.dtype, np.int32)

        def test_noclip(self):
            e = LabyrinthExplore()
            terminated = truncated = False
            e.reset()
            for i in range(5):
                while not (terminated or truncated):
                    _, _, terminated, truncated, _ = e.step(e.action_space.sample())

                visit_mask = e.explored == 1
                obstacle_mask = e.maze.grid == 1
                self.assertFalse(np.any(visit_mask * obstacle_mask == 1))
                e.reset()

        def test_known(self):
            e = LabyrinthExplore((6, 6))
            e.reset(seed=0)
            left, right, up, down = 0, 1, 2, 3
            actions = (
                [left] * 5
                + [up] * 2
                + [right] * 2
                + [left] * 2
                + [down] * 2
                + [right] * 4
                + [up] * 4
                + [left] * 4
            )
            terminated = truncated = False
            cum_rew = 0
            for action in actions:
                self.assertFalse(terminated or truncated)
                obs, reward, terminated, truncated, info = e.step(action)
                cum_rew += reward

            self.assertTrue(terminated)
            self.assertFalse(truncated)
            expected = 1.0 + 8 * e.neg_reward_scale
            self.assertAlmostEqual(cum_rew, expected)



from typing import Optional, Tuple

import numpy as np
from gymnasium import spaces
from gymnasium.core import ActType, ObsType

from popgym.core.env import POPGymEnv
from popgym.core.observability import OBS, STATE
from popgym.core.wrapper import POPGymWrapper

PREV_ACTION = "prev_action"


class PreviousAction(POPGymWrapper):
    """Wrapper that adds the last action to the observation.

    Args:
        env: The environment
        null_action: Optional null action that is returned when resetting the
            environment. If not provided, the null action will be 0
            (int or vector) if it is in the action space, or the lowest action
            possible.

    Returns:
        A gym environment
    """

    def __init__(self, env: POPGymEnv, null_action: Optional[ActType] = None):
        super().__init__(env)
        self.observation_space = PreviousAction.add_act_space_to_obs_space(
            self.env.observation_space, self.env.action_space
        )
        if null_action is None:
            null_action = PreviousAction.get_null_action(self.action_space)
        assert self.action_space.contains(null_action)
        self.null_action = null_action

    @staticmethod
    def add_act_space_to_obs_space(
        observation_space: spaces.Space, action_space: spaces.Space
    ) -> spaces.Space:
        """
        Returns a modified observation space to account for the last action.
        Args:
            observation_space: Original observation space
            action_space: Action space

        Returns:
            The new observation space
        """
        if isinstance(
            observation_space,
            (spaces.Box, spaces.Discrete, spaces.MultiDiscrete, spaces.MultiBinary),
        ):
            observation_space = spaces.Tuple((observation_space, action_space))
        elif isinstance(observation_space, spaces.Tuple):
            observation_space = spaces.Tuple(
                tuple(observation_space.spaces) + (action_space,)
            )
        elif isinstance(observation_space, spaces.Dict):
            observation_space = observation_space.spaces.copy()
            if set(observation_space.keys()) == {OBS, STATE}:
                # Observation comes from ObservabilityWrapper with
                # observability level FULL_AND_PARTIAL
                observation_space[OBS] = PreviousAction.add_act_space_to_obs_space(
                    observation_space[OBS], action_space
                )
            else:
                observation_space[PREV_ACTION] = action_space
            observation_space = spaces.Dict(observation_space)
        else:
            raise NotImplementedError("Unknown observation space")
        return observation_space

    @staticmethod
    def add_act_to_obs(
        observation_space: spaces.Space, obs: ObsType, action: ActType
    ) -> ObsType:
        """
        Static method that adds the action to the observation.
        Args:
            observation_space: Original observation space of the environment.
            obs: The observation.
            action: The action.

        Returns:
            Modified observation.
        """
        if isinstance(
            observation_space,
            (spaces.Box, spaces.Discrete, spaces.MultiDiscrete, spaces.MultiBinary),
        ):
            obs = (obs, action)
        elif isinstance(observation_space, spaces.Tuple):
            obs = (*obs, action)
        elif isinstance(observation_space, spaces.Dict):
            if set(observation_space.keys()) == {OBS, STATE}:
                # Observation comes from ObservabilityWrapper with
                # observability level FULL_AND_PARTIAL
                obs[OBS] = PreviousAction.add_act_to_obs(
                    observation_space[OBS], obs[OBS], action
                )
            else:
                obs[PREV_ACTION] = action
        else:
            raise NotImplementedError("Unknown observation space")
        return obs

    @staticmethod
    def get_null_action(action_space: spaces.Space) -> ActType:
        """
        Static method that generates a null action based on the action space.
        Args:
            action_space: The action space.

        Returns:
            The null action.
        """
        if isinstance(
            action_space,
            (spaces.Discrete, spaces.MultiBinary, spaces.MultiDiscrete, spaces.Box),
        ):
            action = np.zeros(action_space.shape, action_space.dtype)
            if not action_space.contains(action):
                action = action_space.low
        elif isinstance(action_space, spaces.Tuple):
            action = tuple(
                PreviousAction.get_null_action(action_space_)
                for action_space_ in action_space
            )
        elif isinstance(action_space, spaces.Dict):
            action = {
                key: PreviousAction.get_null_action(value)
                for key, value in action_space.items()
            }
        else:
            raise NotImplementedError
        return action

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        obs, reward, terminated, truncated, info = self.env.step(action)
        obs = PreviousAction.add_act_to_obs(self.env.observation_space, obs, action)
        return obs, reward, terminated, truncated, info

    def reset(self, **kwargs):
        obs, info = self.env.reset(**kwargs)
        obs = PreviousAction.add_act_to_obs(
            self.env.observation_space, obs, self.null_action
        )
        return obs, info



from typing import List, Tuple

import gymnasium as gym
import torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType

from popgym.baselines.models.s4d import S4 as S4Model
from popgym.baselines.ray_models.base_model import BaseModel


class S4D(BaseModel):
    r"""The diagonal version of Structured State Spaces (S4) from

    .. code-block:: text

        @inproceedings{gu_efficiently_2022,
            title = {Efficiently Modeling Long Sequences with Structured State Spaces},
            url = {https://openreview.net/forum?id=uYLFoz1vlAC},
            language = {en},
            urldate = {2022-09-22},
            author = {Gu, Albert and Goel, Karan and Re, Christopher},
            month = mar,
            year = {2022},
        }
    """

    MODEL_CONFIG = {
        "embedding": None,
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        self.h = round(self.cfg["hidden_size"] ** 0.5 / 2) * 2
        self.map = torch.nn.Linear(
            self.cfg["preprocessor_output_size"], self.cfg["hidden_size"]
        )
        self.core = S4Model(
            d_model=self.cfg["hidden_size"],
            d_state=self.h,
            mode="diag",
            measure="diag-lin",
            bidirectional=False,
            disc="zoh",
            real_type="exp",
            transposed=False,
        )
        self.did_setup = False
        # Required to initialize some parts of S4

    def initial_state(self) -> List[TensorType]:
        state = self.core.default_state(1).squeeze(0)
        # Ray cannot handle complex floats
        real, imag = state.real, state.imag
        # For each batch, each tensor has shape [1, self.h, self.h]
        return [real, imag]

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:

        # Calling setup_step in init causes deepcopy to fail
        if not self.did_setup:
            self.core.setup_step()
            self.did_setup = True
        real, imag = state
        memory = torch.complex(real, imag)
        z = self.map(z)
        # Inference mode
        if torch.torch.all(seq_lens == 1):
            z, memory = self.core.step(z.squeeze(1), memory)
            z = z.unsqueeze(1)
        # Train mode
        else:
            z, _ = self.core(z)

        return z, [memory.real, memory.imag]



"""The card game Concentration, where the player matches facedown cards"""
import math
from typing import Any, Dict, List, Optional, Tuple

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType

from popgym.core.deck import Deck
from popgym.core.env import POPGymEnv


class Concentration(POPGymEnv):
    """Classic game of concentration. A deck of cards is shuffled and placed
    face-down. The player can flip two cards, if they match they get a reward
    otherwise they dont.

    Args:
        num_decks: Number of decks to play with
        deck_type: String denoting what we are matching. Can be the card colors
        (colors) or the card ranks (ranks)

    Returns:
        A gym environment
    """

    obs_requires_prev_action = True

    def __init__(self, num_decks=1, deck_type="ranks"):
        # See https://math.stackexchange.com/questions/1876467/
        # how-many-turns-on-average-does-it-take-for-a-perfect-player
        # -to-win-concentrati
        self.deck = Deck(num_decks=num_decks)
        n = 52 * num_decks
        self.episode_length = math.ceil(2 * n - (n / (2 * n - 1)))
        self.success_reward_scale = 1 / (self.deck.num_cards // 2)
        self.failure_reward_scale = -1 / (self.episode_length)

        if deck_type == "colors":
            self.deck_type = self.deck.colors
            self.deck_idx_type = self.deck.colors_idx
        # Cannot do suits because there are 13 of each suit -- we need
        # an even number to match
        elif deck_type == "ranks":
            self.deck_type = self.deck.ranks
            self.deck_idx_type = self.deck.ranks_idx
        else:
            raise NotImplementedError(f"Invalid deck type {deck_type}")

        # cards = 14 * np.ones(len(self.deck))
        self.facedown_card = len(np.unique(self.deck_type))
        cards = (1 + self.facedown_card) * np.ones(self.deck.num_cards)
        self.observation_space = gym.spaces.MultiDiscrete(cards)
        self.state_space = gym.spaces.Tuple(
            (
                gym.spaces.MultiDiscrete(cards - 1),  # cards values
                gym.spaces.MultiBinary(self.deck.num_cards),  # cards faces on
                gym.spaces.Discrete(n),  # first card turned
                gym.spaces.Discrete(n),  # second card turned
            )
        )
        self.action_space = gym.spaces.Discrete(self.deck.num_cards)
        self.deck.add_players("face_up", "face_up_idx", "in_play", "in_play_idx")
        self.last_in_play_idx = []
        self.last_trying_card_already_up = False

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        reward = 0
        terminated = truncated = False

        # Done conditions
        if self.curr_step >= self.episode_length - 1:
            truncated = True

        # Flip card
        self.deck["in_play"].append(self.deck.idx[action])
        self.deck["in_play_idx"].append(action)

        assert len(self.deck["in_play"]) <= 2
        # Cannot flip the same card twice
        self.obs = self.get_obs()
        self.last_in_play_idx = self.deck["in_play_idx"].copy()

        assert len(self.deck["face_up"]) == len(self.deck["face_up_idx"])
        assert len(self.deck["in_play"]) == len(self.deck["in_play_idx"])

        trying_card_already_up = any(
            idx in self.deck["face_up_idx"] for idx in self.deck["in_play_idx"]
        )

        # End of phase
        if trying_card_already_up:
            reward = self.failure_reward_scale * len(self.deck["in_play_idx"])
            self.deck.discard_hands("in_play", "in_play_idx")
        elif len(self.deck["in_play"]) == 2:
            inplay_cards = self.deck_type[self.deck["in_play"]]
            cards_match = inplay_cards[0] == inplay_cards[1]
            flipped_same_idx = self.deck["in_play"][0] == self.deck["in_play"][1]
            if cards_match and not flipped_same_idx:
                reward = self.success_reward_scale
                self.deck["face_up"].extend(self.deck["in_play"])
                self.deck["face_up_idx"].extend(self.deck["in_play_idx"])
                terminated = len(self.deck["face_up"]) == len(self.deck)
            else:
                reward = 2 * self.failure_reward_scale

            # Flip two last flipped-up cards face down again
            self.deck.discard_hands("in_play", "in_play_idx")
            self.hand: List[int] = []

        self.curr_step += 1

        return self.obs, reward, terminated, truncated, {}

    def render(self, mode="ascii"):
        self.obs.tolist()
        visible_mask = self.obs != self.facedown_card
        rend = np.full(self.obs.shape, "?")
        rend[visible_mask] = self.obs[visible_mask]
        rend = rend.reshape(4, 13)
        output = (
            " "
            + str(rend)
            .replace("[", "")
            .replace("]", "")
            .replace(",", "")
            .replace("'", "")
            # .replace("\n", "")
        )
        print(output)

    def get_obs(self):
        obs = self.facedown_card * np.ones(len(self.deck), dtype=np.int64)
        hand_idx = self.deck["face_up_idx"] + self.deck["in_play_idx"]
        deck_idx = self.deck.idx[hand_idx]
        obs[hand_idx] = self.deck_idx_type[deck_idx]
        return obs

    def get_state(self):
        cards_face_up = np.zeros(
            len(
                self.deck,
            ),
            dtype=np.int8,
        )
        cards_face_up[self.deck["face_up_idx"]] = 1
        first_card, second_card = self.facedown_card, self.facedown_card
        if len(self.last_in_play_idx) == 2:
            first_card, second_card = self.last_in_play_idx
        elif len(self.last_in_play_idx) == 1:
            first_card = self.last_in_play_idx[0]
        return self.state.copy(), cards_face_up, first_card, second_card

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[gym.core.ObsType, Dict[str, Any]]:
        super().reset(seed=seed)
        self.flip_next_turn = False
        self.deck.reset(rng=self.np_random)
        self.state = self.deck_idx_type[self.deck.idx].copy().astype(np.float32)
        self.curr_step = 0
        self.turn_counter = 0
        self.flipped_counter = 0
        self.last_in_play_idx = []
        self.last_trying_card_already_up = False
        self.obs = self.get_obs()
        info: Dict[str, Any] = {}
        return self.obs, info


class ConcentrationEasy(Concentration):
    def __init__(self):
        super().__init__(num_decks=1, deck_type="colors")


class ConcentrationMedium(Concentration):
    def __init__(self):
        super().__init__(num_decks=2, deck_type="colors")


class ConcentrationHard(Concentration):
    def __init__(self):
        super().__init__(num_decks=1, deck_type="ranks")



from typing import Optional, Tuple

import torch
from torch import nn


class IndRNN(nn.Module):
    def __init__(
        self, input_size, hidden_size, activation="relu", clamp=True, max_len=1024
    ):
        super().__init__()
        self.clamp = clamp
        self.clamp_val = 2 ** (1 / max_len)
        self.input0 = nn.Linear(input_size, hidden_size)
        self.weight0 = nn.Parameter(torch.empty(hidden_size))
        self.bias0 = nn.Parameter(torch.empty(hidden_size))

        self.input1 = nn.Linear(hidden_size, hidden_size)
        self.weight1 = nn.Parameter(torch.empty(hidden_size))
        self.bias1 = nn.Parameter(torch.empty(hidden_size))

        self.output = nn.Linear(hidden_size, hidden_size)

        torch.nn.init.uniform_(self.weight0, 0, 1)
        torch.nn.init.uniform_(self.weight1, 0, 1)
        torch.nn.init.uniform_(self.bias0, 0, 1)
        torch.nn.init.uniform_(self.bias1, 0, 1)

    def clamp_weights(self) -> None:
        self.weight0.data.clamp_(-self.clamp_val, self.clamp_val)
        self.weight1.data.clamp_(-self.clamp_val, self.clamp_val)

    def forward(
        self, x: torch.Tensor, state: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        input:
            x: [B, T, F]
            state: [B, 2, F]
        output:
            y: [B, T, D]
            state: [B, 2, D]
        """
        if state is None:
            state = torch.zeros(x.shape[0], 2, *self.weight0.shape, device=x.device)

        state0, state1 = state.chunk(2, dim=1)
        B, T, _ = x.shape
        if self.clamp:
            self.clamp_weights()
        x = self.input0(x)
        outs = []
        for t in range(T):
            state0 = torch.relu(x[:, t, None] + state0 * self.weight0 + self.bias0)
            state1 = torch.relu(
                self.input1(state0) + state1 * self.weight1 + self.bias1
            )
            outs.append(state1)

        outs = torch.cat(outs, dim=1)
        return outs, torch.stack([state0[:, -1], state1[:, -1]], dim=1)



from typing import List, Tuple

import torch
from torch import nn

from popgym.baselines.models.aggregations import get_aggregator


class Phi(nn.Module):
    def forward(self, x):
        return torch.nn.functional.elu(x) + 1


class LinearAttentionBlock(nn.Module):
    """
    The building block from the Linear Transformers are Secretly RNNs Paper. This is
    a form of linear transformer.

    Inputs:
        input_size: Size of input feature dim
        hidden_size: Size of key/query/value space
        S_aggregator: Which type of aggregation to use for the numerator (S term)
        Z_aggregator: Which type of aggregation to use for the denominator (Z term)
    """

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        S_aggregator: str = "sum",
        Z_aggregator: str = "sum",
    ):
        super().__init__()
        self.key = nn.Linear(input_size, hidden_size, bias=False)
        self.query = nn.Linear(input_size, hidden_size, bias=False)
        self.value = nn.Linear(input_size, hidden_size, bias=False)
        self.norm = nn.LayerNorm(input_size)
        self.phi = Phi()
        self.S_aggregator = get_aggregator(S_aggregator)()
        self.Z_aggregator = get_aggregator(Z_aggregator)()

    def forward(
        self, x: torch.Tensor, state: List[torch.Tensor]
    ) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """
        Input:
            x: [B, T, F]
            state: Tuple[
                [B, 1, D, D],
                [B, 1, D]
            ]
        Output:
            y: [B, T, D]
            state: Tuple[
                [B, 1, D, D],
                [B, 1, D]
            ]
        """

        x = self.norm(x)
        K = self.phi(self.key(x))
        Q = self.phi(self.query(x))
        V = self.value(x)
        S, Z = state
        B, T, F = K.shape

        # S = sum(K V^T)
        S = self.S_aggregator(
            torch.einsum("bti, btj -> btij", K, V).reshape(B, T, F * F),
            S.reshape(B, 1, F * F),
        ).reshape(B, T, F, F)
        # Z = sum(K)
        Z = self.Z_aggregator(K, Z.reshape(B, 1, F))
        # numerator = Q^T S
        numerator = torch.einsum("bti, btil -> btl", Q, S)
        # denominator = Q^T Z
        denominator = torch.einsum("bti, btl -> bt", Q, Z).reshape(B, T, 1) + 1e-5
        # output = (Q^T S) / (Q^T Z)
        output = numerator / denominator

        state = [S, Z]

        return output, state



import unittest

from popgym.envs import has_mazelib

if has_mazelib():
    from popgym.envs.labyrinth_escape import LabyrinthEscape

    class TestLabyrinthEscape(unittest.TestCase):
        def test_step(self):
            e = LabyrinthEscape()
            e.reset()
            for i in range(1000):
                _, _, terminated, truncated, _ = e.step(e.action_space.sample())
                if terminated or truncated:
                    e.reset()

        def test_tostring(self):
            e = LabyrinthEscape((6, 6))
            e.reset()
            e.tostring()

        def test_goal(self):
            e = LabyrinthEscape((6, 6))
            terminated = truncated = False
            e.reset()
            for i in range(5):
                while not (terminated or truncated):
                    _, _, terminated, truncated, _ = e.step(e.action_space.sample())
                self.assertTrue(e.curr_step < e.max_episode_length)
                e.reset()

        def test_known(self):
            e = LabyrinthEscape((8, 8))
            e.reset(seed=1)
            left, right, _, down = 0, 1, 2, 3
            actions = [down] * 5 + [left] * 2 + [down] * 2 + [right] * 6 + [down] * 1
            done = False
            cum_rew = 0
            for action in actions:
                self.assertFalse(done)
                obs, reward, terminated, truncated, info = e.step(action)
                cum_rew += reward

            self.assertTrue(terminated)
            self.assertFalse(truncated)
            expected = 1.0 + len(actions) * e.neg_reward_scale
            self.assertAlmostEqual(cum_rew, expected)



# Inspired by ray rllib at
# https://github.com/ray-project/ray/blob/master/rllib/examples/env/stateless_cartpole.py

import math
from typing import Optional, Tuple

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType
from gymnasium.envs.classic_control import CartPoleEnv
from gymnasium.spaces import Box

from popgym.core.env import POPGymEnv


class StatelessCartPole(CartPoleEnv, POPGymEnv):
    """Partially observable variant of the CartPole gym environment.

    https://github.com/openai/gym/blob/master/gym/envs/classic_control/
    cartpole.py
    We delete the x- and angular velocity components of the state, so that it
    can only be solved by a memory enhanced model (policy).

    Args:
        max_episode_length: Exactly what it sounds like
    Returns:
        A gym environment
    """

    def __init__(self, *args, **kwargs):
        self.max_episode_length = kwargs.pop("max_episode_length", 200)
        super().__init__(*args, **kwargs)

        # Fix our observation-space (remove 2 velocity components).
        high = np.array(
            [
                self.x_threshold * 2,
                self.theta_threshold_radians * 2,
            ],
            dtype=np.float32,
        )
        self.state_space = self.observation_space
        self.observation_space = Box(low=-high, high=high, dtype=np.float32)

    def get_state(self):
        state = np.array(self.state, dtype=np.float32)
        return state

    def reward_transform(self, reward):
        if math.isclose(reward, 0):
            transformed = -1.0
        else:
            transformed = 1.0 / self.max_episode_length
        return transformed

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        next_obs, reward, terminated, truncated, info = super().step(action)
        self.num_steps += 1
        if self.num_steps >= self.max_episode_length:
            # Agent beat episode
            truncated = True
        reward = self.reward_transform(reward)
        # next_obs is [x-pos, x-veloc, angle, angle-veloc]
        next_obs = np.array([next_obs[0], next_obs[2]])
        return next_obs, reward, terminated, truncated, info

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[dict] = None
    ) -> Tuple[gym.core.ObsType, dict]:
        self.num_steps = 0
        init_obs, info = super().reset(seed=seed, options=options)
        init_obs = np.array([init_obs[0], init_obs[2]])
        return init_obs, info


class StatelessCartPoleEasy(StatelessCartPole):
    pass


class StatelessCartPoleMedium(StatelessCartPole):
    def __init__(self, *args, **kwargs):
        super().__init__(max_episode_length=400)


class StatelessCartPoleHard(StatelessCartPole):
    def __init__(self, *args, **kwargs):
        super().__init__(max_episode_length=600)



import math
from typing import List, Tuple

import gymnasium as gym
import torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType
from torch import nn

from popgym.baselines.models.fwp import FWPBlock
from popgym.baselines.ray_models.base_model import BaseModel


class FastWeightProgrammer(BaseModel):
    r"""The fast weight programmer from

    .. code-block:: text

        @inproceedings{schlag_linear_2021,
            title = {
                Linear {Transformers} {Are} {Secretly} {Fast} {Weight} {Programmers}
            },
            url = {https://proceedings.mlr.press/v139/schlag21a.html},
            language = {en},
            urldate = {2022-09-21},
            booktitle = {
                Proceedings of the 38th International Conference on Machine Learning
            },
            publisher = {PMLR},
            author = {Schlag, Imanol and Irie, Kazuki and Schmidhuber, Jürgen},
            month = jul,
            year = {2021},
            note = {ISSN: 2640-3498},
            pages = {9355--9366},
        }

    without the RNN extensions."""

    MODEL_CONFIG = {
        # Whether to use the sum normalization over the key/query term
        # as in the paper
        "sum_normalization": True,
        # Which positional embedding to use
        "embedding": "sine",
        # Which cumulative aggregator to use. Only sum is used in the paper.
        # This can be sum or max
        "aggregator": "sum",
        # Whether the recurrent state size should be hidden_size
        # or hidden_size ** 2
        "big_hidden": False,
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        if self.cfg["big_hidden"]:
            self.h = self.cfg["hidden_size"]
        else:
            self.h = round(math.sqrt(self.cfg["hidden_size"]))
        self.core = FWPBlock(
            input_size=self.cfg["preprocessor_output_size"],
            hidden_size=self.h,
            aggregator=self.cfg["aggregator"],
            sum_normalization=self.cfg["sum_normalization"],
        )
        if not self.cfg["big_hidden"]:
            self.unmap = nn.Linear(self.h, self.cfg["hidden_size"])

    def initial_state(self) -> List[TensorType]:
        return [
            torch.zeros(1, self.h, self.h),
        ]

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:
        B, T, _ = z.shape
        [memory] = state
        z, memory = self.core(z, memory)
        h = z.shape[-1]
        if not self.cfg["big_hidden"]:
            z = self.unmap(z)
        return z, [
            memory[:, -1].reshape(B, 1, h, h),
        ]


class BigFastWeightProgrammer(FastWeightProgrammer):
    MODEL_CONFIG = {
        # Whether to use the sum normalization over the key/query term
        # as in the paper
        "sum_normalization": True,
        # Which positional embedding to use
        "embedding": "sine",
        # Which cumulative aggregator to use. Only sum is used in the paper.
        # This can be sum or max
        "aggregator": "sum",
        # Whether the recurrent state size should be hidden_size
        # or hidden_size ** 2
        "big_hidden": True,
    }



"""Collection of POPGym environments
"""

import inspect
from importlib import find_loader
from typing import Any, Dict

import gymnasium as gym

from popgym.envs.autoencode import (
    Autoencode,
    AutoencodeEasy,
    AutoencodeHard,
    AutoencodeMedium,
)
from popgym.envs.battleship import (
    Battleship,
    BattleshipEasy,
    BattleshipHard,
    BattleshipMedium,
)
from popgym.envs.concentration import (
    Concentration,
    ConcentrationEasy,
    ConcentrationHard,
    ConcentrationMedium,
)
from popgym.envs.count_recall import (
    CountRecall,
    CountRecallEasy,
    CountRecallHard,
    CountRecallMedium,
)
from popgym.envs.higher_lower import (
    HigherLower,
    HigherLowerEasy,
    HigherLowerHard,
    HigherLowerMedium,
)
from popgym.envs.minesweeper import (
    MineSweeper,
    MineSweeperEasy,
    MineSweeperHard,
    MineSweeperMedium,
)
from popgym.envs.multiarmed_bandit import (
    MultiarmedBandit,
    MultiarmedBanditEasy,
    MultiarmedBanditHard,
    MultiarmedBanditMedium,
)
from popgym.envs.noisy_stateless_cartpole import (
    NoisyStatelessCartPole,
    NoisyStatelessCartPoleEasy,
    NoisyStatelessCartPoleHard,
    NoisyStatelessCartPoleMedium,
)
from popgym.envs.noisy_stateless_pendulum import (
    NoisyStatelessPendulum,
    NoisyStatelessPendulumEasy,
    NoisyStatelessPendulumHard,
    NoisyStatelessPendulumMedium,
)
from popgym.envs.repeat_first import (
    RepeatFirst,
    RepeatFirstEasy,
    RepeatFirstHard,
    RepeatFirstMedium,
)
from popgym.envs.repeat_previous import (
    RepeatPrevious,
    RepeatPreviousEasy,
    RepeatPreviousHard,
    RepeatPreviousMedium,
)
from popgym.envs.stateless_cartpole import (
    StatelessCartPole,
    StatelessCartPoleEasy,
    StatelessCartPoleHard,
    StatelessCartPoleMedium,
)
from popgym.envs.stateless_pendulum import (
    StatelessPendulum,
    StatelessPendulumEasy,
    StatelessPendulumHard,
    StatelessPendulumMedium,
)

#
# Simple envs
#
DIAGNOSTIC: Dict[gym.Env, Dict[str, Any]] = {
    Autoencode: {"id": "popgym-Autoencode-v0"},
    RepeatPrevious: {"id": "popgym-RepeatPrevious-v0"},
    RepeatFirst: {"id": "popgym-RepeatFirst-v0"},
    CountRecall: {"id": "popgym-CountRecall-v0"},
}

DIAGNOSTIC_EASY: Dict[gym.Env, Dict[str, Any]] = {
    AutoencodeEasy: {"id": "popgym-AutoencodeEasy-v0"},
    RepeatPreviousEasy: {"id": "popgym-RepeatPreviousEasy-v0"},
    RepeatFirstEasy: {"id": "popgym-RepeatFirstEasy-v0"},
    CountRecallEasy: {"id": "popgym-CountRecallEasy-v0"},
}

DIAGNOSTIC_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
    AutoencodeMedium: {"id": "popgym-AutoencodeMedium-v0"},
    RepeatPreviousMedium: {"id": "popgym-RepeatPreviousMedium-v0"},
    RepeatFirstMedium: {"id": "popgym-RepeatFirstMedium-v0"},
    CountRecallMedium: {"id": "popgym-CountRecallMedium-v0"},
}

DIAGNOSTIC_HARD: Dict[gym.Env, Dict[str, Any]] = {
    AutoencodeHard: {"id": "popgym-AutoencodeHard-v0"},
    RepeatPreviousHard: {"id": "popgym-RepeatPreviousHard-v0"},
    RepeatFirstHard: {"id": "popgym-RepeatFirstHard-v0"},
    CountRecallHard: {"id": "popgym-CountRecallHard-v0"},
}

ALL_DIAGNOSTIC = {**DIAGNOSTIC_EASY, **DIAGNOSTIC_MEDIUM, **DIAGNOSTIC_HARD}

#
# Control envs
#
CONTROL: Dict[gym.Env, Dict[str, Any]] = {
    StatelessCartPole: {"id": "popgym-StatelessCartPole-v0"},
    StatelessPendulum: {
        "id": "popgym-StatelessPendulum-v0",
    },
    # BipedalWalker: {"id": "popgym-BipedalWalker-v0"},
}

CONTROL_EASY: Dict[gym.Env, Dict[str, Any]] = {
    StatelessCartPoleEasy: {"id": "popgym-StatelessCartPoleEasy-v0"},
    StatelessPendulumEasy: {
        "id": "popgym-StatelessPendulumEasy-v0",
    },
    # BipedalWalkerEasy: {"id": "popgym-BipedalWalkerEasy-v0"},
}

CONTROL_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
    StatelessCartPoleMedium: {"id": "popgym-StatelessCartPoleMedium-v0"},
    StatelessPendulumMedium: {
        "id": "popgym-StatelessPendulumMedium-v0",
    },
    # BipedalWalkerMedium: {"id": "popgym-BipedalWalkerMedium-v0"},
}

CONTROL_HARD: Dict[gym.Env, Dict[str, Any]] = {
    StatelessCartPoleHard: {"id": "popgym-StatelessCartPoleHard-v0"},
    StatelessPendulumHard: {
        "id": "popgym-StatelessPendulumHard-v0",
    },
    # BipedalWalkerHard: {"id": "popgym-BipedalWalkerHard-v0"},
}

ALL_CONTROL = {**CONTROL_EASY, **CONTROL_MEDIUM, **CONTROL_HARD}

#
# Noisy envs
#
NOISY: Dict[gym.Env, Dict[str, Any]] = {
    NoisyStatelessCartPole: {"id": "popgym-NoisyStatelessCartPole-v0"},
    NoisyStatelessPendulum: {
        "id": "popgym-NoisyStatelessPendulum-v0",
    },
    MultiarmedBandit: {"id": "popgym-MultiarmedBandit-v0"},
    # NonstationaryBandit: {"id": "popgym-NonstationaryBandit-v0"},
}

NOISY_EASY: Dict[gym.Env, Dict[str, Any]] = {
    NoisyStatelessCartPoleEasy: {"id": "popgym-NoisyStatelessCartPoleEasy-v0"},
    NoisyStatelessPendulumEasy: {
        "id": "popgym-NoisyStatelessPendulumEasy-v0",
    },
    MultiarmedBanditEasy: {"id": "popgym-MultiarmedBanditEasy-v0"},
    # NonstationaryBanditEasy: {"id": "popgym-NonstationaryBanditEasy-v0"},
}

NOISY_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
    NoisyStatelessCartPoleMedium: {"id": "popgym-NoisyStatelessCartPoleMedium-v0"},
    NoisyStatelessPendulumMedium: {
        "id": "popgym-NoisyStatelessPendulumMedium-v0",
    },
    MultiarmedBanditMedium: {"id": "popgym-MultiarmedBanditMedium-v0"},
    # NonstationaryBanditMedium: {"id": "popgym-NonstationaryBanditMedium-v0"},
}

NOISY_HARD: Dict[gym.Env, Dict[str, Any]] = {
    NoisyStatelessCartPoleHard: {"id": "popgym-NoisyStatelessCartPoleHard-v0"},
    NoisyStatelessPendulumHard: {
        "id": "popgym-NoisyStatelessPendulumHard-v0",
    },
    MultiarmedBanditHard: {"id": "popgym-MultiarmedBanditHard-v0"},
    # NonstationaryBanditHard: {"id": "popgym-NonstationaryBanditHard-v0"},
}

ALL_NOISY = {**NOISY_EASY, **NOISY_MEDIUM, **NOISY_HARD}

#
# Game envs
#
GAME: Dict[gym.Env, Dict[str, Any]] = {
    HigherLower: {"id": "popgym-HigherLower-v0"},
    Battleship: {"id": "popgym-Battleship-v0"},
    Concentration: {"id": "popgym-Concentration-v0"},
    MineSweeper: {"id": "popgym-MineSweeper-v0"},
}

GAME_EASY: Dict[gym.Env, Dict[str, Any]] = {
    HigherLowerEasy: {"id": "popgym-HigherLowerEasy-v0"},
    BattleshipEasy: {"id": "popgym-BattleshipEasy-v0"},
    ConcentrationEasy: {"id": "popgym-ConcentrationEasy-v0"},
    MineSweeperEasy: {"id": "popgym-MineSweeperEasy-v0"},
}

GAME_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
    HigherLowerMedium: {"id": "popgym-HigherLowerMedium-v0"},
    BattleshipMedium: {"id": "popgym-BattleshipMedium-v0"},
    ConcentrationMedium: {"id": "popgym-ConcentrationMedium-v0"},
    MineSweeperMedium: {"id": "popgym-MineSweeperMedium-v0"},
}

GAME_HARD: Dict[gym.Env, Dict[str, Any]] = {
    HigherLowerHard: {"id": "popgym-HigherLowerHard-v0"},
    BattleshipHard: {"id": "popgym-BattleshipHard-v0"},
    ConcentrationHard: {"id": "popgym-ConcentrationHard-v0"},
    MineSweeperHard: {"id": "popgym-MineSweeperHard-v0"},
}

ALL_GAME = {**GAME_EASY, **GAME_MEDIUM, **GAME_HARD}

#
# Navigation envs
#
def has_mazelib():  # noqa: E302
    """Check if mazelib is installed"""
    return find_loader("mazelib") is not None


if has_mazelib():
    # mazelib can somtimes be a headache to install
    # mazes are also a poor test of memory
    from popgym.envs.labyrinth_escape import (
        LabyrinthEscape,
        LabyrinthEscapeEasy,
        LabyrinthEscapeHard,
        LabyrinthEscapeMedium,
    )
    from popgym.envs.labyrinth_explore import (
        LabyrinthExplore,
        LabyrinthExploreEasy,
        LabyrinthExploreHard,
        LabyrinthExploreMedium,
    )

    NAVIGATION: Dict[gym.Env, Dict[str, Any]] = {
        LabyrinthExplore: {"id": "popgym-LabyrinthExplore-v0"},
        LabyrinthEscape: {"id": "popgym-LabyrinthEscape-v0"},
    }

    NAVIGATION_EASY: Dict[gym.Env, Dict[str, Any]] = {
        LabyrinthExploreEasy: {"id": "popgym-LabyrinthExploreEasy-v0"},
        LabyrinthEscapeEasy: {"id": "popgym-LabyrinthEscapeEasy-v0"},
    }

    NAVIGATION_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
        LabyrinthExploreMedium: {"id": "popgym-LabyrinthExploreMedium-v0"},
        LabyrinthEscapeMedium: {"id": "popgym-LabyrinthEscapeMedium-v0"},
    }

    NAVIGATION_HARD: Dict[gym.Env, Dict[str, Any]] = {
        LabyrinthExploreHard: {"id": "popgym-LabyrinthExploreHard-v0"},
        LabyrinthEscapeHard: {"id": "popgym-LabyrinthEscapeHard-v0"},
    }
else:
    NAVIGATION = {}
    NAVIGATION_EASY = {}
    NAVIGATION_MEDIUM = {}
    NAVIGATION_HARD = {}

ALL_NAVIGATION = {
    **NAVIGATION_EASY,
    **NAVIGATION_MEDIUM,
    **NAVIGATION_HARD,
}

#
# Sets of envs
#
ALL_BASE: Dict[gym.Env, Dict[str, Any]] = {
    **DIAGNOSTIC,
    **CONTROL,
    **NOISY,
    **GAME,
    **NAVIGATION,
}
ALL_EASY: Dict[gym.Env, Dict[str, Any]] = {
    **DIAGNOSTIC_EASY,
    **CONTROL_EASY,
    **NOISY_EASY,
    **GAME_EASY,
    **NAVIGATION_EASY,
}
ALL_MEDIUM: Dict[gym.Env, Dict[str, Any]] = {
    **DIAGNOSTIC_MEDIUM,
    **CONTROL_MEDIUM,
    **NOISY_MEDIUM,
    **GAME_MEDIUM,
    **NAVIGATION_MEDIUM,
}
ALL_HARD: Dict[gym.Env, Dict[str, Any]] = {
    **DIAGNOSTIC_HARD,
    **CONTROL_HARD,
    **NOISY_HARD,
    **GAME_HARD,
    **NAVIGATION_HARD,
}
ALL: Dict[gym.Env, Dict[str, Any]] = {
    **ALL_EASY,
    **ALL_MEDIUM,
    **ALL_HARD,
}

# Register envs
for e, v in ALL.items():
    mod_name = inspect.getmodule(e).__name__  # type: ignore
    gym.envs.register(
        entry_point=":".join([mod_name, e.__name__]), order_enforce=False, **v
    )
# For Sphinx automodule docs
del e



"""Guess if the next card in the deck is higher or lower than the last card drawn."""
from typing import Any, Dict, Optional, Tuple

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType

from popgym.core.deck import RANKS, Deck
from popgym.core.env import POPGymEnv


def value_fn(hand):
    if hand[-1] > hand[-2]:
        return 1
    elif hand[-1] == hand[-2]:
        return 0
    else:
        return -1


class HigherLower(POPGymEnv):
    """A game of higher/lower

    Given a deck of cards, the agent predicts whether the
    next card drawn from the deck is higher or lower than the last card drawn from
    the deck. A push results in zero reward, while a correct/incorrect guess result
    in 1/deck_size and -1/deck_size reward. The agent can learn to count cards to
    infer which cards are left in the deck, improving accuracy.

    Args:
        num_decks: The number of individual decks combined into a single deck.

    Returns:
        A gym environment
    """

    def __init__(self, num_decks=1):
        self.num_decks = num_decks
        self.deck = Deck(num_decks)
        self.deck.add_players("player")
        self.action_space = gym.spaces.Discrete(2)
        self.state = np.zeros(
            (
                len(
                    RANKS,
                )
            ),
            dtype=np.uint8,
        )
        self.observation_space = self.deck.get_obs_space(["ranks"])
        self.state_space = gym.spaces.Box(0, 1, self.state.shape)
        self.value_map = dict(zip(self.deck.ranks, range(len(self.deck.ranks))))
        self.deck_size = len(self.deck)

    def get_state(self):
        return (self.state.copy() / 4 / self.num_decks).astype(np.float32)

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        guess_higher = action == 0
        terminated = len(self.deck) <= 1

        self.deck.deal("player", 1)
        assert self.deck.hand_size("player") == 2
        curr_value, next_value = self.deck.show("player", ["ranks_idx"]).reshape(-1)
        self.state[next_value] += 1

        rew_scale = 1 / self.deck_size
        if next_value == curr_value:
            reward = 0
        elif (next_value > curr_value) == guess_higher:
            reward = rew_scale
        else:
            reward = -rew_scale

        self.deck.discard("player", 0)
        obs = self.deck.show("player", ["ranks_idx"]).item()

        return obs, reward, terminated, False, {}

    def render(self, mode="ascii"):
        return self.deck.visualize_idx(self.deck["player"])

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[gym.core.ObsType, Dict[str, Any]]:
        super().reset(seed=seed)
        self.deck.reset(rng=self.np_random)
        self.deck.deal("player", 1)
        self.state = np.zeros(
            (
                len(
                    RANKS,
                )
            ),
            dtype=np.uint8,
        )
        obs = self.deck.show("player", ["ranks_idx"]).item()
        self.state[obs] = 1
        viz = np.concatenate(self.deck.show("player", ["suits", "ranks"]))
        return obs, {"card": viz}


class HigherLowerEasy(HigherLower):
    def __init__(self):
        super().__init__(num_decks=1)


class HigherLowerMedium(HigherLower):
    def __init__(self):
        super().__init__(num_decks=2)


class HigherLowerHard(HigherLower):
    def __init__(self):
        super().__init__(num_decks=3)



"""Receive some cards then output how many of each you've seen"""
from typing import Any, Dict, Optional, Tuple

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType

from popgym.core.deck import Deck
from popgym.core.env import POPGymEnv


class CountRecall(POPGymEnv):
    """A game where the agent is queried on past events

    A game where the agent is queried on how many times it has observed
    an event in the past. This tests long-term order-agnostic memory like sets.

    Args:
        max_episode_length: The maximum number of timesteps in an episode
        error_clamp: Denotes the domain of the linear portion of the reward
            function. E.g. error_clamp == 2 means the agent will receive
            linearly-decreasing rewards for counts off by up to 2. Errors greater
            than 2 provide the same negative reward as an error of 2. Note as
            error_clamp -> inf, a crappy agent and perfect agent will acheive
            the same total reward.
        deck_type: What we use to count/differentiate cards.
            Can be colors, suits, or ranks.

    Returns:
        A gym environment
    """

    query: Optional[int]

    def __init__(self, num_decks=1, error_clamp=0.5, deck_type="colors"):
        self.value_deck = Deck(num_decks=num_decks)
        self.query_deck = Deck(num_decks=num_decks)
        self.value_deck.add_players("in_play")
        self.query_deck.add_players("in_play")
        if deck_type == "colors":
            self.deck_type = self.value_deck.colors
            self.deck_idx_type = self.value_deck.colors_idx
        elif deck_type == "suits":
            self.deck_type = self.value_deck.suits
            self.deck_idx_type = self.value_deck.suits_idx
        elif deck_type == "ranks":
            self.deck_type = self.value_deck.ranks
            self.deck_idx_type = self.value_deck.ranks_idx
        else:
            raise NotImplementedError(f"Invalid deck type {deck_type}")

        self.num_distinct_cards = len(np.unique(self.deck_type))
        self.max_card_count = int(self.value_deck.num_cards / self.num_distinct_cards)
        # Space: [dealt card, card query]
        self.observation_space = gym.spaces.MultiDiscrete([self.num_distinct_cards] * 2)
        self.state_space = gym.spaces.Tuple(
            (
                gym.spaces.Box(0.0, 1.0, (self.num_distinct_cards,)),
                gym.spaces.Box(0.0, 1.0, (self.num_distinct_cards,)),
                # The current query is needed, but not the dealt card
                self.observation_space,
            )
        )
        self.last_obs = np.array([0, 0])

        self.action_space = gym.spaces.Discrete(self.max_card_count)
        self.max_episode_length = self.value_deck.num_cards - 1
        self.error_clamp = error_clamp
        self.reward_scale = 1 / self.max_episode_length

    def render(self):
        dealt = self.deck_type[self.value]
        queried = self.deck_type[self.query]
        print(f"Dealt {dealt}, recall {queried} count")

    def get_state(self):
        state = (
            (self.counts / self.max_card_count).astype(np.float32),
            (self.query_counts / self.max_card_count).astype(np.float32),
            self.last_obs.copy(),
        )
        return state

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        if isinstance(action, np.ndarray):
            action = action.item()

        self.prev_query = self.query
        prev_count = self.counts[self.prev_query]
        self.value, self.query = self.deal()
        self.counts[self.value] += 1
        self.query_counts[self.query] += 1

        reward = 1 if action == prev_count else -1
        reward *= self.reward_scale

        terminated = len(self.value_deck) == 0

        obs = np.array([self.value, self.query], dtype=np.int64)
        self.last_obs = obs
        info = {"counts": self.counts}

        return obs.copy(), reward, terminated, False, info

    def sample_deck(self):
        return np.random.choice(self.deck_idx_type, size=self.max_episode_length + 1)

    def deal(self):
        self.value_deck.deal("in_play")
        self.query_deck.deal("in_play")
        return (
            self.deck_idx_type[self.value_deck["in_play"][-1]],
            self.deck_idx_type[self.query_deck["in_play"][-1]],
        )

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[gym.core.ObsType, Dict[str, Any]]:
        super().reset(seed=seed)
        self.value_deck.reset(rng=self.np_random)
        self.query_deck.reset(rng=self.np_random)  # Having another PRNG is not needed.

        self.counts = np.zeros((self.num_distinct_cards,), dtype=np.int64)
        self.query_counts = np.zeros((self.num_distinct_cards,), dtype=np.int64)

        self.value, self.query = self.deal()
        self.prev_query = None
        self.counts[self.value] += 1
        self.query_counts[self.query] += 1

        obs = np.array([self.value, self.query], dtype=np.int64)
        self.last_obs = obs.copy()
        return obs, {"counts": self.counts}


class CountRecallEasy(CountRecall):
    pass


class CountRecallMedium(CountRecall):
    def __init__(self, *args, **kwargs):
        super().__init__(num_decks=2, deck_type="suits")


class CountRecallHard(CountRecall):
    def __init__(self, *args, **kwargs):
        super().__init__(num_decks=4, deck_type="ranks")



from typing import List, Tuple

import gymnasium as gym
import torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType
from torch import nn

from popgym.baselines.ray_models.base_model import BaseModel


class GRU(BaseModel):
    r"""The gated recurrent unit from

    .. code-block:: text

        @article{chung_empirical_2014,
            title = {
                Empirical evaluation of gated recurrent neural
                networks on sequence modeling
            },
            journal = {arXiv preprint arXiv:1412.3555},
            author = {
                Chung, Junyoung and Gulcehre, Caglar and Cho,
                KyungHyun and Bengio, Yoshua
            },
            year = {2014},
        }
    """

    MODEL_CONFIG = {
        # Number of recurrent hidden layers in encoder/decoder
        "num_recurrent_layers": 1,
        "benchmark": False,
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        # Need to define self.core
        if self.cfg["benchmark"]:
            self.core = nn.GRUCell(
                self.cfg["preprocessor_output_size"],
                self.cfg["hidden_size"],
            )
        else:
            self.core = nn.GRU(
                self.cfg["preprocessor_output_size"],
                self.cfg["hidden_size"],
                self.cfg["num_recurrent_layers"],
                batch_first=True,
            )

    def initial_state(self) -> List[TensorType]:
        return [torch.zeros(1, self.cfg["hidden_size"])]

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:

        memory = state[0].permute(1, 0, 2)

        if self.cfg["benchmark"]:
            outs = []
            memory = memory.squeeze(0)
            for t in range(z.shape[1]):
                memory = self.core(z[:, t], memory)
                outs.append(memory)
            z = torch.stack(outs, dim=1)
            memory = memory.unsqueeze(0)
        else:
            z, memory = self.core(z, memory)
        # State expected to be list
        state = [memory.permute(1, 0, 2)]  # type: ignore

        return z, state



import math

from popgym.envs.higher_lower import HigherLower
from tests.base_env_test import AbstractTest


class TestHigherLower(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = HigherLower()

    def test_episode_higher(self):
        terminated = truncated = False
        o, _ = self.env.reset()
        obs_list = [o]
        reward_list = []
        decklen = 52
        while not (terminated or truncated):
            obs, reward, terminated, truncated, info = self.env.step(0)
            obs_list.append(obs)
            reward_list.append(reward)
        pairs = list(zip(obs_list[:-1], obs_list[1:]))
        pred_correct = [j > i for i, j in pairs]
        pred_incorrect = [j < i for i, j in pairs]
        pred_rew = (sum(pred_correct) - sum(pred_incorrect)) / decklen
        actual_rew = sum(reward_list)

        self.assertTrue(math.isclose(pred_rew, actual_rew))



from collections import OrderedDict
from typing import Dict, List, Tuple, Union

import gymnasium as gym
import torch
from dnc import DNC
from ray.rllib.utils.typing import ModelConfigDict, TensorType

from popgym.baselines.ray_models.base_model import BaseModel


class DiffNC(BaseModel):
    r"""The differentiable neural computer from

    .. code-block:: text

        @techreport{wayne_unsupervised_2018,
            title = {Unsupervised {Predictive} {Memory} in a {Goal}-{Directed} {Agent}},
            url = {http://arxiv.org/abs/1803.10760},
            number = {arXiv:1803.10760},
            urldate = {2022-09-09},
            institution = {arXiv},
            month = mar,
            year = {2018},
            doi = {10.48550/arXiv.1803.10760},
        }
    """

    MODEL_CONFIG = {
        # Number of controller hidden layers
        "num_hidden_layers": 1,
        # Number of LSTM units
        "num_layers": 1,
        # Number of read heads, i.e. how many addrs are read at once
        "read_heads": 4,
        # Size of each cell, also
        # number of cells == hidden_size // cell_size
        "cell_size": 16,
        # LSTM activation function
        "nonlinearity": "tanh",
    }

    MEMORY_KEYS = [
        "memory",
        "link_matrix",
        "precedence",
        "read_weights",
        "write_weights",
        "usage_vector",
    ]

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        self.cfg["nr_cells"] = self.cfg["hidden_size"] // self.cfg["cell_size"]
        self.dnc_built = False

    def initial_state(self) -> List[TensorType]:
        ctrl_hidden = [
            torch.zeros(self.cfg["num_hidden_layers"], self.cfg["hidden_size"]),
            torch.zeros(self.cfg["num_hidden_layers"], self.cfg["hidden_size"]),
        ]
        m = self.cfg["nr_cells"]
        r = self.cfg["read_heads"]
        w = self.cfg["cell_size"]
        memory = [
            torch.zeros(m, w),  # memory
            torch.zeros(1, m, m),  # link_matrix
            torch.zeros(1, m),  # precedence
            torch.zeros(r, m),  # read_weights
            torch.zeros(1, m),  # write_weights
            torch.zeros(m),  # usage_vector
        ]

        read_vecs = torch.zeros(w * r)

        state = [*ctrl_hidden, read_vecs, *memory]
        return state

    def unpack_state(
        self,
        state: List[TensorType],
    ) -> Tuple[List[Tuple[TensorType, TensorType]], Dict[str, TensorType], TensorType]:
        """Given a list of tensors, reformat for self.dnc input"""
        assert len(state) == 9, "Failed to verify unpacked state"
        ctrl_hidden: List[Tuple[TensorType, TensorType]] = [
            (
                state[0].permute(1, 0, 2).contiguous(),
                state[1].permute(1, 0, 2).contiguous(),
            )
        ]
        read_vecs: TensorType = state[2]
        memory: List[TensorType] = state[3:]
        memory_dict: OrderedDict[str, TensorType] = OrderedDict(
            zip(self.MEMORY_KEYS, memory)
        )

        return ctrl_hidden, memory_dict, read_vecs

    def pack_state(
        self,
        ctrl_hidden: List[Tuple[TensorType, TensorType]],
        memory_dict: Dict[str, TensorType],
        read_vecs: TensorType,
    ) -> List[TensorType]:
        """Given the dnc output, pack it into a list of tensors
        for rllib state. Order is ctrl_hidden, read_vecs, memory_dict"""
        state = []
        ctrl_hidden = [
            ctrl_hidden[0][0].permute(1, 0, 2),
            ctrl_hidden[0][1].permute(1, 0, 2),
        ]
        state += ctrl_hidden
        assert len(state) == 2, "Failed to verify packed state"
        state.append(read_vecs)
        assert len(state) == 3, "Failed to verify packed state"
        state += memory_dict.values()
        assert len(state) == 9, "Failed to verify packed state"
        return state

    def build_dnc(self, device_idx: Union[int, None]) -> torch.nn.Module:
        return DNC(
            input_size=self.cfg["preprocessor_input_size"],
            hidden_size=self.cfg["hidden_size"],
            num_layers=self.cfg["num_layers"],
            num_hidden_layers=self.cfg["num_hidden_layers"],
            read_heads=self.cfg["read_heads"],
            cell_size=self.cfg["cell_size"],
            nr_cells=self.cfg["nr_cells"],
            nonlinearity=self.cfg["nonlinearity"],
            gpu_id=device_idx,
        )

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:
        B, T, F = z.shape

        # First run
        if not self.dnc_built:
            gpu_id = z.device.index if z.device.index is not None else -1
            self.core = self.build_dnc(gpu_id)
            self.core.output = torch.nn.Linear(
                self.core.output.in_features, self.cfg["hidden_size"]
            ).to(z.device)
            self.dnc_built = True

        unpacked = self.unpack_state(state)
        output, unpacked = self.core(z, unpacked)
        packed = self.pack_state(*unpacked)

        return output, packed
